/*
Various utility functions for the validation code
This file is in the public domain
*/

#include <iostream>
#include <string>
#include <typeinfo>

#include "filters.h"
#include "encoder.h"
#include "symkey.h"
#include "modes.h"
using namespace OpenCL_types;

// XXX: stupid nasty hack... Why isn't there something like
// this (but better) in the library? We don't have to be using a pipe
// at all, it's quite silly...
OpenCL::SymmetricKey decode(const std::string& hex_in)
   {
   OpenCL::Pipe decoder(new OpenCL::HexDecoder);
   decoder.write((byte*)hex_in.c_str(), hex_in.size());
   decoder.close();
   u32bit read = decoder.remaining();
   return OpenCL::SymmetricKey(decoder.read_all(), read);
   }

OpenCL::Filter* get_block(const std::string&, const OpenCL::BlockCipherKey&);
OpenCL::Filter* get_mode(const std::string&, const OpenCL::BlockCipherKey&,
                 const OpenCL::BlockCipherModeIV&);
OpenCL::Filter* get_stream(const std::string&, const OpenCL::StreamCipherKey&);
OpenCL::Filter* get_hash(const std::string&);
OpenCL::Filter* get_mac(const std::string&, const OpenCL::MACKey&);

OpenCL::Filter* lookup(const std::string& algname, const std::string& hex_key,
               const std::string& hex_iv)
   {
   OpenCL::SymmetricKey key = decode(hex_key);
   OpenCL::BlockCipherModeIV iv = decode(hex_iv);

   OpenCL::Filter* filter;

   filter = get_block(algname, key);
   if(filter) return filter;

   filter = get_mode(algname, key, iv);
   if(filter) return filter;

   filter = get_stream(algname, key);
   if(filter) return filter;

   filter = get_hash(algname);
   if(filter) return filter;

   filter = get_mac(algname, key);
   if(filter) return filter;

   return 0;
   }
