#!/usr/bin/perl -w

#############################TODO########################################
# What I really should do is a write a script that generates this script on the
# basis of some description files (for each OS, CPU, and CC type). I could even
# have it print shell instead of Perl (maybe), so people wouldn't need to have
# Perl installed to build stuff.

# Compilers: AIX CCs, Intel CC (x86), MetroWerks (x86 + PowerPC)

# CPUs: POWER (AIX cc + GCC)
# OSes: NetBSD, OpenBSD, AIX, SunOS 4, MacOS

#############################README######################################
# The point of this script is to generate some basic makefile options
# which depend heavily on the compiler, OS, and arch used. Basically, the
# goal is to generate optimal options for the specified system. This
# requires the script know crazy stuff like what machine-level optimization
# options are supported by any particular compiler. The code is ugly and pretty
# much of the brute-force stupidity class.

# Keep in mind that I personally don't have access to many of these machines,
# and thus values were basically guessed based on available documentation. If
# some value causes problems (or if you have docs for a compiler not included
# in here), drop me an email. Also many perverse or unusual cases will probably
# cause it to break. :(

# Jack Lloyd (lloyd@acm.jhu.edu)

#########################################################################
# Name of the file to dump everything to
$CONFIGFILE = "Config";

##########################################################################
# Some tables defining the relationships between architectures, submodel
# types, compilers, and OSes.

@ARCHS=("ia32", "ia64", "mips", "mips64", "sparc", "sparc64", "alpha",
        "arm", "powerpc", "m68k", "hppa");

# We accept these names as aliases for the canonical names given above
%ARCH_ALIASES = (
   ia32      => [ "x86", "ix86", "80x86" ],
   mips      => [ "mipseb", "mipsel" ], # RPM uses these
   alpha     => [ "axp" ],
   powerpc   => [ "ppc" ],
   m68k      => [ "680x0", "68k" ],
   hppa      => [ "hp-pa", "parisc", "pa-risc", "hp-parisc", "hp-pa-risc" ]
);

# Submodels of the CPU type
%SUBMODELS = (
   ia32      => [ "i386", "i486", "i586", "i686" ],
   ia64      => [ "ia64" ],
   mips      => [ "r3000", "r6000" ],
   mips64    => [ "r4000", "r4400", "r5000", "r8000", "r10000" ],
   sparc     => [ "sparcv7", "sparcv8", "sparcv9" ],
   sparc64   => [ "ultra", "ultra2", "ultra2i" ],
   alpha     => [ "ev4", "ev5", "ev56", "pca56", "ev6", "ev67" ],
   arm       => [ "arm2", "arm3", "arm3m", "arm4", "arm4t", "arm5" ],
   powerpc   => [ "ppc601", "ppc602", "ppc603", "ppc604" ],
   m68k      => [ "68020", "68030", "68040", "68060" ],
   hppa      => [ "hppa1.0", "hppa1.1", "hppa2.0" ]
);

# Generally the most basic type is chosen as the default
%SUBMODEL_DEFAULT = (
   ia32      => "i386",
   ia64      => "ia64",
   mips      => "r3000",
   mips64    => "r4400",
   sparc     => "sparcv7",
   sparc64   => "ultra",
   alpha     => "ev4",
   arm       => "arm2",
   powerpc   => "ppc601",
   m68k      => "68020",
   hppa      => "hppa1.0"
);

# Matches compilers to a list of CPUs they support
%CCS_ARCH = (
   gcc       => [ "ia32", "ia64", "mips", "mips64", "sparc", "sparc64",
                  "alpha", "arm", "powerpc", "m68k", "hppa" ],
   kcc       => [ "ia32", "mips", "sparc", "alpha", "hppa" ],
   sunwspro  => [ "ia32", "sparc", "sparc64" ],
   pgi       => [ "ia32" ],
   mipspro   => [ "mips", "mips64" ],
   compaq    => [ "alpha" ],
   hpcc      => [ "hppa" ]
);

# Matches compilers to a list of OSes they support
%CCS_OS = (
   gcc       => [ "linux", "freebsd", "solaris", "cygwin", "irix",
                  "tru64", "hpux", "beos" ],
   kcc       => [ "linux", "solaris", "irix", "tru64", "hpux" ],
   compaq    => [ "linux", "tru64" ],
   pgi       => [ "linux", "solaris" ],
   mipspro   => [ "irix" ],
   sunwspro  => [ "solaris" ],
   hpcc      => [ "hpux" ]
);

# Matches OSes to a list of CPUs they support
%OSES = (
   linux        => [ "ia32", "ia64", "mips", "mips64", "sparc", "sparc64",
                     "alpha", "arm", "powerpc", "m68k", "hppa" ],
   solaris      => [ "ia32", "sparc", "sparc64" ],
   freebsd      => [ "ia32", "alpha" ],
   beos         => [ "ia32", "powerpc" ],
   irix         => [ "mips", "mips64" ],
   tru64        => [ "alpha" ],
   hpux         => [ "hppa" ],
   cygwin       => [ "ia32" ],
);

# Any OSes that support shared objects are listed here - this is an OS/CPU
# tuple, where * means any CPU type - I don't know about BeOS
@OS_SUPPORTS_SHARED_OBJ = (
   "(linux/*)", "(freebsd/*)", "(solaris/*)",
   "(tru64/*)", "(irix/*)", "(hpux/*)"
);

# OSes listed here use GNU ld as their default linker - I'm not sure about BeOS
@OS_USES_GNULD = ("linux", "freebsd", "beos", "cygwin");

##########################################################################
# Map internal names to real world names

%REALNAME_OS = (
   linux        => "Linux",
   freebsd      => "FreeBSD",
   solaris      => "Solaris",
   tru64        => "Tru64",
   irix         => "IRIX",
   hpux         => "HP-UX",
   cygwin       => "Cygwin",
   beos         => "BeOS"
);

%REALNAME_CC = (
   gcc       => "GNU C++",
   kcc       => "KAI C++",
   pgi       => "PGI C++",
   compaq    => "Compaq C++",
   mipspro   => "SGI MIPSPro C++",
   sunwspro  => "Sun Workshop Pro",
   hpcc      => "HP-UX CC"
);

%REALNAME_ARCH = (
   ia32      => "IA-32",
   ia64      => "IA-64",
   mips      => "MIPS",
   mips64    => "64-bit MIPS",
   sparc     => "Sun SPARC",
   sparc64   => "64-bit Sun SPARC",
   alpha     => "DEC Alpha",
   arm       => "ARM",
   powerpc   => "PowerPC",
   m68k      => "Motorola 68K",
   hppa      => "HP Precision"
);

############################################################################

# if we weren't given 1, 2, or 3 arguments
if($#ARGV < 0 or $#ARGV > 2) { print_help(); }

$no_debug = 0;

$parseme = "";

foreach $arg (@ARGV)
   {
   # We lowercase all args mostly to allow a script (like RPM) to directly
   # use the output of 'uname' (or some other predefined thing) for input
   $arg =~ tr/A-Z/a-z/;
   if($arg eq "--help") { print_help(); }
   elsif($arg eq "--no-debug") { $no_debug = 1; }
   else { $parseme = $arg; }
   }

($cc,$os,$arch) = split(/-/,$parseme);
if(!defined($cc) or !defined($os) or !defined($arch)) { print_help(); }
if(!defined($CCS_OS{$cc})) { print "ERROR: Compiler $cc isn't known\n"; exit; }
if(!defined($OSES{$os}) and $os ne "generic")
   { print "ERROR: OS $os isn't known\n"; exit; }

if($arch eq "generic") { $uberarch = "generic"; }

# If it's a generic CPU, skip all this
if($arch ne "generic")
   {
   # First, if arch is an alias, convert it to the canonical name
   foreach $cputype (@ARCHS) {
      foreach $alias (@{ $ARCH_ALIASES{$cputype} }) {
         if($alias eq $arch) { $arch = $cputype; }
         }
      }

   # At the end, we want $arch to be the submodel name, while $uberarch is the
   # generic family name.
   if(defined($SUBMODELS{$arch}))
      { # generic name was given - default to something reasonable
      $uberarch = $arch;
      $arch = $SUBMODEL_DEFAULT{$arch};
      # Prevent warnings about ia64, which has no submodels (yet). We denote
      # this situation by setting the single submodel type to the same name
      # as the name of the arch family. This will be needed for x86-64, too.
      if($arch ne $uberarch)
         {
         print "WARNING: Using $arch as default type for family ",
               $REALNAME_ARCH{$uberarch},"\n";
         }
      }
   else # submodel name - need to figure out what family it's from
      {
      foreach $cputype (@ARCHS) {
         foreach $submodel (@{ $SUBMODELS{$cputype} }) {
            if($submodel eq $arch) { $uberarch = $cputype; }
            }
         }
      }
   }

if(!defined($uberarch)) { print "ERROR: CPU type $arch isn't known\n"; exit; }

# Check to make sure the compiler and arch match up ok
unless(list_to_scalar(@{ $CCS_ARCH{$cc} }) =~ /$uberarch/ or
       $uberarch eq "generic")
   { print "ERROR: Using $REALNAME_CC{$cc} with ".
           "$REALNAME_ARCH{$uberarch} doesn't make sense.\n"; exit; }

# Check to make sure the OS and arch match up ok
unless(list_to_scalar(@{ $OSES{$os} }) =~ /$uberarch/ or
       $uberarch eq "generic" or $os eq "generic")
   { print "ERROR: Using $REALNAME_OS{$os} with $REALNAME_ARCH{$uberarch} ".
           "doesn't make sense.\n"; exit; }

# Check to make sure the compiler and OS match up ok
unless(list_to_scalar(@{ $CCS_OS{$cc} }) =~ /$os/ or $os eq "generic")
   { print "ERROR: Using $REALNAME_CC{$cc} with $REALNAME_OS{$os} doesn't ".
           "make sense.\n"; exit; }

@results = let_the_evil_begin($cc,$no_debug,$os,$arch,$uberarch);

print_output(@results);

exit;

# Some nice innocuous functions
sub print_output
   {
   my($CXX,$OPT_FLAGS,$MACH_FLAGS,$LANG_FLAGS,$WARN_FLAGS,$SO_OBJ_FLAGS,
      $SO_LINK_FLAGS,$MAKE_SHARED) = @_;
   open CONFIGFILE, ">$CONFIGFILE" or die "Couldn't open output file ($!)\n";

   print CONFIGFILE<<END_CONFIG;
CXX           = $CXX
OPT_FLAGS     = $OPT_FLAGS
MACH_FLAGS    = $MACH_FLAGS
LANG_FLAGS    = $LANG_FLAGS
WARN_FLAGS    = $WARN_FLAGS
END_CONFIG

   if($MAKE_SHARED)
      {
      print CONFIGFILE<<END_SO_INFO;
MAKE_SHARED   = yes
SO_OBJ_FLAGS  = $SO_OBJ_FLAGS
SO_LINK_FLAGS = $SO_LINK_FLAGS
END_SO_INFO
      }
   else
      {
      print CONFIGFILE "MAKE_SHARED   = no\n";
      }
   }

sub print_help
   {
   print <<"   ENDOFHELP";
Usage: $0 [--no-debug] <compiler>-<os>-<cpu>

Use --no-debug to configure the compiler to enable further optimizations which
may prevent debugging on some systems.

You may use 'generic' for OS and CPU, which will disable any specializations
that would ordinarily be made (useful if your OS/CPU aren't listed).

   ENDOFHELP
   print "Known compilers: "; print list_to_scalar(sort keys %CCS_ARCH);
   print "Known OSes: "; print list_to_scalar(sort keys %OSES);

   print "\n";
   foreach $arch (@ARCHS)
      { print "Known '$arch' models: ";
        print list_to_scalar(@{ $SUBMODELS{$arch} }); }

   exit;
   }

sub list_to_scalar
   {
   my $data = "";
   foreach $elem (@_)
      {
      if($elem ne $_[0]) { $data .= " "; }
      $data .= $elem;
      }
   $data .= "\n";
   return $data;
   }

#############################################################################
# This code is gross. It also rots quickly, as it depends on the
# implementation of many different compilers and OSes. :(
sub let_the_evil_begin
   {
   my($cc,$no_debug,$os,$arch,$uberarch) = @_;

   $MAKE_SHARED = 0;
   # This will never be true for generic, so we don't have a special case
   if(list_to_scalar(@OS_SUPPORTS_SHARED_OBJ) =~ m#$os/$uberarch# or
      list_to_scalar(@OS_SUPPORTS_SHARED_OBJ) =~ m#$os/\*#)
      { $MAKE_SHARED = 1; }

   # CXX, OPT_FLAGS, LANG_FLAGS, WARN_FLAGS, and probably SO_OBJ_FLAGS
   # shouldn't be CPU or OS dependant, so we just need to take the compiler
   # into account.
   if($cc eq "gcc")
      { $CXX = "g++";
        $OPT_FLAGS = "-O3 -fstrict-aliasing";
        $LANG_FLAGS = "-ansi -pedantic";
        $WARN_FLAGS = "-W -Wall";
        $SO_OBJ_FLAGS = "-fPIC"; }

   if($cc eq "kcc")
      { $CXX = "KCC";
        $OPT_FLAGS = "+K3 --inline_auto_space_time=65 --abstract_pointer";
        $LANG_FLAGS = "-D__KAI_STRICT --strict";
        $WARN_FLAGS = "";
        $SO_OBJ_FLAGS = ""; }

   if($cc eq "pgi")
      { $CXX = "pgCC";
        $OPT_FLAGS = "-fast -Minline";
        $LANG_FLAGS = "";
        $WARN_FLAGS = "";
        $SO_OBJ_FLAGS = "-fPIC"; }

   if($cc eq "mipspro")
      { $CXX = "CC";
        $OPT_FLAGS = "-O3 -OPT:alias=TYPED";
        $LANG_FLAGS = "-ansi -LANG:ansi-for-init-scope=ON";
        $WARN_FLAGS = "-fullwarn";
        $SO_OBJ_FLAGS = "-KPIC"; }

   if($cc eq "sunwspro")
      { $CXX = "CC";
        $OPT_FLAGS = "-fast";
        $LANG_FLAGS = "";
        $WARN_FLAGS = "+w2";
        $SO_OBJ_FLAGS = "-KPIC"; }

   if($cc eq "compaq")
      { $CXX = "cxx";
        $OPT_FLAGS = "-fast";
        $LANG_FLAGS = "-std ansi";
        $WARN_FLAGS = "";
        $SO_OBJ_FLAGS = ""; }

   # I don't have access to an HP-SUX machine at the moment, so this is
   # total speculation (based off HP's online man pages).
   if($cc eq "hpcc")
      { $CXX = "CC";
        $OPT_FLAGS = "+O4 +Oinlinebudget=150";
        $LANG_FLAGS = "-ext +eh -z";
        $WARN_FLAGS = "+w";
        $SO_OBJ_FLAGS = "+Z"; }

   if($MAKE_SHARED)
      {
      # What to give a compiler to tell it to make a shared lib
      %CXX_SHARED_CMD = (
         gcc        => "-shared -fPIC",
         mipspro    => "-shared",
         sunwspro   => "-G",
         compaq     => "-shared",
         hpcc       => "-b",

         # Talk about brain damaged. Same compiler, same CPU, mutually
         # unsupported flags for creating a shared lib. <sigh>
         pgi_linux   => "-shared",
         pgi_solaris => "-G"
      );

      # How to set the SONAME
      %SONAME_CMD = (
         # Most CC's have their own command
         kcc      => "--soname \$(SONAME)", # KCC's method
         sunwspro => "-h\$(SONAME)", # SunWSPro's method
         compaq   => "-soname \$(SONAME)", # Compaq's method
         hpcc     => "-Wl,+h,\$(SONAME)", # Not actually documented in CC(1) (?)

         pgi_linux => "-Wl,-soname,\$(SONAME)",
         pgi_solaris => "-Wl,-h,\$(SONAME)",

         # GCC doesn't, so various versions of it go here
         gnu       => "-Wl,-soname,\$(SONAME)",
         solaris   => "-mimpure-text -Wl,-h,\$(SONAME)",
         tru64     => "-Wl,-soname,\$(SONAME)",
         irix      => "-Wl,-soname,\$(SONAME)",
         hpux      => "-Wl,+h,\$(SONAME)"
                     );

      if($cc eq "pgi")
         {
         unless($os eq "linux" or $os eq "solaris")
            {
            print "Sorry, Portland Group C++ is brain damaged, and we don't\n",
                  "know what to do here. Bailing out...\n";
            exit;
            }
         $cc = "pgi_".$os;
         }

      # If the C++ compiler has it's own soname command, use that
      if(defined($SONAME_CMD{$cc})) { $ld = $cc; }
      elsif($cc eq "gcc" and list_to_scalar(@OS_USES_GNULD) =~ /$os/)
         { $ld = "gnu" }
      else { $ld = $os; }

      if(defined($CXX_SHARED_CMD{$cc}) && defined($SONAME_CMD{$ld}))
         { $SO_LINK_FLAGS = $CXX_SHARED_CMD{$cc} . " " . $SONAME_CMD{$ld}; }
      # presumably the compiler figures it out somehow without us telling it
      elsif(defined($CXX_SHARED_CMD{$cc}))
         { $SO_LINK_FLAGS = $CXX_SHARED_CMD{$cc}; }
      elsif(defined($SONAME_CMD{$ld}))
         { $SO_LINK_FLAGS = $SONAME_CMD{$ld}; }
      }

   # now it gets _really_ messy
   $MACH_FLAGS = ""; # default to none - perfectly ok, just not as fast

   if($arch ne "generic")
   {
      if($no_debug)
         {
         if($cc eq "gcc") { $MACH_FLAGS = "-fomit-frame-pointer "; }
         }

      if($cc eq "gcc")
         {
         if($uberarch eq "ia32")
            {
            if($arch eq "i386") { $MACH_FLAGS .= "-mcpu=i486"; }
            else { $MACH_FLAGS .= "-march=$arch"; }
            }
         elsif($uberarch eq "mips")
            {
            if($arch eq "r3000") { $isa = 1; }
            elsif($arch eq "r6000") { $isa = 2; }
            else { print "ERROR: MIPS CPU type $arch is legal but no specs ",
                         "given for GCC.\n"; exit 1; }
            $CXX .= " -mabi=n32"; # Needed for linking too
            $MACH_FLAGS .= "-mcpu=$arch -mips$isa";
            }
         elsif($uberarch eq "mips64")
            {
            if($arch eq "r4000") { $isa = 3; }
            elsif($arch eq "r4400") { $isa = 3; }
            elsif($arch eq "r5000") { $isa = 4; }
            elsif($arch eq "r8000") { $isa = 4; }
            elsif($arch eq "r10000") { $arch = "r8000"; $isa = 4; }
            else { print "ERROR: MIPS CPU type $arch is legal but no specs ",
                         "given for GCC.\n"; exit 1; }
            $CXX .= " -mabi=64"; # Needed for linking too
            $MACH_FLAGS .= "-mcpu=$arch -mips$isa";
            }
         elsif($uberarch eq "sparc")
            { 
            $MACH_FLAGS .= "-mcpu=$arch ";
            if($arch eq "v9") { $MACH_FLAGS .= "-m32"; } # Force 32-bit code
            }
         elsif($uberarch eq "sparc64")
            { 
            $MACH_FLAGS .= "-mcpu=v9 -m64 "; # Ummm... 64 bits. :)
            }
         elsif($uberarch eq "alpha")
            { if($arch eq "ev67") { $arch = "ev6"; }
               $MACH_FLAGS .= "-mcpu=$arch"; }
         elsif($uberarch eq "arm")
            { $arch =~ s/v//; $MACH_FLAGS .= "-march=$arch"; }
         elsif($uberarch eq "powerpc")
            { $arch =~ s/^ppc//; $MACH_FLAGS .= "-mcpu=$arch";  }
         elsif($uberarch eq "m68k") { $MACH_FLAGS .= "-m$arch"; }
         elsif($uberarch eq "hppa")
            { $arch =~ s/^hppa//; $MACH_FLAGS .= "-march=$arch"; }
         }

      elsif($cc eq "mipspro")
         {
         if($uberarch eq "mips")
            {
            if($arch eq "r3000") { $isa = 1; }
            elsif($arch eq "r6000") { $isa = 2; }
            else { print "ERROR: MIPS CPU type $arch is legal but no specs ",
                         "given for GCC.\n"; exit 1; }
            $CXX .= " -n32"; # Needed for linking too
            $MACH_FLAGS .= "-mcpu=$arch -mips$isa";
            }
         elsif($uberarch eq "mips64")
            {
            if($arch eq "r4000") { $isa = 3; }
            elsif($arch eq "r4400") { $isa = 3; }
            elsif($arch eq "r5000") { $isa = 4; }
            elsif($arch eq "r8000") { $isa = 4; }
            elsif($arch eq "r10000") { $arch = "r8000"; $isa = 4; }
            else { print "ERROR: MIPS CPU type $arch is legal but no specs ",
                         "given for GCC.\n"; exit 1; }
            $CXX .= " -64"; # Needed for linking too
            $MACH_FLAGS .= "-mips$isa ";
            # MIPSPro has special optimizers for these guys
            if($arch eq "r5000" or $arch eq "r8000" or $arch eq "r10000")
               { $MACH_FLAGS .= "-$arch"; }
            }
         }
      elsif($cc eq "sunwspro")
         {
         if($uberarch eq "ia32")
            {
            if($arch eq "i386") { $MACH_FLAGS .= "-xchip=486 -xarch=386"; }
            if($arch eq "i486") { $MACH_FLAGS .= "-xchip=486 -xarch=486"; }
            if($arch eq "i586")
               { $MACH_FLAGS .= "-xchip=pentium -xarch=pentium"; }
            if($arch eq "i686")
               { $MACH_FLAGS .= "-xchip=pentium_pro -xarch=pentium_pro"; }
            }
         elsif($uberarch eq "sparc")
            {
            $MACH_FLAGS .= "-xtarget=generic ";
            if($arch eq "sparcv7") { $MACH_FLAGS .= "-xarch=v7"; }
            if($arch eq "sparcv8") { $MACH_FLAGS .= "-xarch=v8"; }
            # -xarch=v9 makes 64 bit code, which we don't want. So use v8
            if($arch eq "sparcv9") { $MACH_FLAGS .= "-xchip=ultra -xarch=v8"; }
            }
         elsif($uberarch eq "sparc64")
            {
            $MACH_FLAGS .= "-xtarget=generic ";
            if($arch eq "ultra") { $MACH_FLAGS .= "-xchip=ultra -xarch=v9"; }
            if($arch eq "ultra2") { $MACH_FLAGS .= "-xchip=ultra2 -xarch=v9"; }
            if($arch eq "ultra2i") { $MACH_FLAGS .= "-xchip=ultra2i -xarch=v9"; }
            }
         }
      elsif($cc eq "compaq")
         {
         $MACH_FLAGS .= "-arch=$arch";
         }
      elsif($cc eq "hpcc")
         {
         if($arch eq "hppa1.1") { $MACH_FLAGS .= "+DA1.1"; }
         if($arch eq "hppa2.0") { $MACH_FLAGS .= "+DA2.0"; }
         }
      elsif($cc =~ /^pgi/)
         {
         if($uberarch eq "ia32")
            {
            if($arch ne "i586" and $arch ne "i686")
               { $MACH_FLAGS .= "-tp px"; }
            if($arch eq "i586") { $MACH_FLAGS .= "-tp p5"; }
            if($arch eq "i686") { $MACH_FLAGS .= "-tp p6 "; }
            }
         }

      # strip trailing spaces
      $MACH_FLAGS =~ s/ *$//;
   }

   return ($CXX,$OPT_FLAGS,$MACH_FLAGS,$LANG_FLAGS,$WARN_FLAGS,
           $SO_OBJ_FLAGS,$SO_LINK_FLAGS,$MAKE_SHARED);
   }

##########################################################################
