/*************************************************
* CS-Cipher Source File                          *
* (C) 1999-2001 The OpenCL Project               *
*************************************************/

#include "cscipher.h"

namespace OpenCL {

/*************************************************
* CS-Cipher Encryption                           *
*************************************************/
void CS_Cipher::encrypt(const byte in[BLOCKSIZE], byte out[BLOCKSIZE]) const
   {
   SecureBuffer<byte, 8> buffer;
   xor_buf(out, in, round_key, 8);
   for(u32bit j = 0; j != 24; j++)
      {
      buffer[0] = P[TE1[out[0]] ^ out[1]];
      buffer[4] = P[ROT[out[0]] ^ out[1]];
      buffer[1] = P[TE1[out[2]] ^ out[3]];
      buffer[5] = P[ROT[out[2]] ^ out[3]];
      buffer[2] = P[TE1[out[4]] ^ out[5]];
      buffer[6] = P[ROT[out[4]] ^ out[5]];
      buffer[3] = P[TE1[out[6]] ^ out[7]];
      buffer[7] = P[ROT[out[6]] ^ out[7]];
      xor_buf(out, buffer, round_key + 8*j + 8, 8);
      }
   }

/*************************************************
* CS-Cipher Decryption                           *
*************************************************/
void CS_Cipher::decrypt(const byte in[BLOCKSIZE], byte out[BLOCKSIZE]) const
   {
   SecureBuffer<byte, 8> buffer;
   xor_buf(out, in, round_key + 192, 8);
   for(u32bit j = 0; j != 24; j++)
      {
      buffer[0] = TD1[ P[out[0]] ^ P[out[4]]];
      buffer[1] = ROT[buffer[0]] ^ P[out[4]];
      buffer[2] = TD1[ P[out[1]] ^ P[out[5]]];
      buffer[3] = ROT[buffer[2]] ^ P[out[5]];
      buffer[4] = TD1[ P[out[2]] ^ P[out[6]]];
      buffer[5] = ROT[buffer[4]] ^ P[out[6]];
      buffer[6] = TD1[ P[out[3]] ^ P[out[7]]];
      buffer[7] = ROT[buffer[6]] ^ P[out[7]];
      xor_buf(out, buffer, round_key + 184 - 8*j, 8);
      }
   }

/*************************************************
* CS-Cipher Key Setup                            *
*************************************************/
void CS_Cipher::set_key(const byte key[], u32bit len) throw(InvalidKeyLength)
   {
   if(len != KEYLENGTH)
      throw InvalidKeyLength("CS-Cipher");
   static const byte C[8] = { 0xB7, 0xE1, 0x51, 0x62, 0x8A, 0xED, 0x2A, 0x6A };
   static const byte D[8] = { 0xBF, 0x71, 0x58, 0x80, 0x9C, 0xF4, 0xF3, 0xC7 };
   SecureBuffer<byte, 8> buffer, matrix, K[11];
   K[0].copy(key + 8, 8);
   K[1].copy(key, 8);
   for(u32bit j = 2; j != 11; j++)
      {
      xor_buf(buffer, K[j-1], P + 8*(j-2), 8);
      for(u32bit k = 0; k != 8; k++)
         buffer[k] = P[buffer[k]];
      matrix.clear();
      for(u32bit k = 0, mask = 0x80; k != 8; k++, mask >>= 1)
         for(u32bit l = 0; l != 8; l++)
            matrix[k] |= (k >= l) ? ((buffer[l] & mask) << (k-l)) :
                                    ((buffer[l] & mask) >> (l-k));
      xor_buf(K[j], matrix, K[j-2], 8);
      }
   for(u32bit j = 0; j != 192; j += 24)
      {
      round_key.copy(j, K[j/24 + 2], 8);
      round_key.copy(j +  8, C, 8);
      round_key.copy(j + 16, D, 8);
      }
   round_key.copy(192, K[10], 8);
   }

}
