/*************************************************
* Skipjack Source File                           *
* (C) 1999-2001 The OpenCL Project               *
*************************************************/

#include "skipjack.h"

namespace OpenCL {

/*************************************************
* Skipjack Encryption                            *
*************************************************/
void Skipjack::encrypt(const byte in[BLOCKSIZE], byte out[BLOCKSIZE]) const
   {
   u16bit W1 = make_u16bit(in[0], in[1]), W2 = make_u16bit(in[2], in[3]);
   u16bit W3 = make_u16bit(in[4], in[5]), W4 = make_u16bit(in[6], in[7]);
   for(u32bit j = 0; j != 2; j++)
      {
      for(u32bit k = 0; k != 8; k++)
         {
         byte G1 = get_byte(0, W1); byte G2 = get_byte(1, W1);
         byte G3 = F[G2 ^ K[(4*(16*j+k)  )%10]] ^ G1;
         G1 = F[G3 ^ K[(4*(16*j+k)+1)%10]] ^ G2;
         G2 = F[G1 ^ K[(4*(16*j+k)+2)%10]] ^ G3;
         G3 = F[G2 ^ K[(4*(16*j+k)+3)%10]] ^ G1;
         u16bit T = make_u16bit(G2, G3);
         W1 = (u16bit)(T ^ (16*j+k+1) ^ W4);
         W4 = W3; W3 = W2; W2 = T;
         }
      for(u32bit k = 8; k != 16; k++)
         {
         u16bit T = W1; W1 = W4; W4 = W3;
         W3 = (u16bit)(W2 ^ (16*j+k+1) ^ T);
         byte G1 = get_byte(0, T); byte G2 = get_byte(1, T);
         byte G3 = F[G2 ^ K[(4*(16*j+k)  )%10]] ^ G1;
         G1 = F[G3 ^ K[(4*(16*j+k)+1)%10]] ^ G2;
         G2 = F[G1 ^ K[(4*(16*j+k)+2)%10]] ^ G3;
         G3 = F[G2 ^ K[(4*(16*j+k)+3)%10]] ^ G1;
         W2 = make_u16bit(G2, G3);
         }
      }
   out[0] = get_byte(0, W1); out[1] = get_byte(1, W1);
   out[2] = get_byte(0, W2); out[3] = get_byte(1, W2);
   out[4] = get_byte(0, W3); out[5] = get_byte(1, W3);
   out[6] = get_byte(0, W4); out[7] = get_byte(1, W4);
   }

/*************************************************
* Skipjack Decryption                            *
*************************************************/
void Skipjack::decrypt(const byte in[BLOCKSIZE], byte out[BLOCKSIZE]) const
   {
   u16bit W1 = make_u16bit(in[0], in[1]), W2 = make_u16bit(in[2], in[3]);
   u16bit W3 = make_u16bit(in[4], in[5]), W4 = make_u16bit(in[6], in[7]);
   for(s32bit j = 1; j >= 0; j--)
      {
      for(u32bit k = 15; k >= 8; k--)
         {
         byte G1 = get_byte(1, W2); byte G2 = get_byte(0, W2);
         byte G3 = F[G2 ^ K[(4*(16*j+k)+3)%10]] ^ G1;
         G1 = F[G3 ^ K[(4*(16*j+k)+2)%10]] ^ G2;
         G2 = F[G1 ^ K[(4*(16*j+k)+1)%10]] ^ G3;
         G3 = F[G2 ^ K[(4*(16*j+k)  )%10]] ^ G1;
         u16bit T = make_u16bit(G3, G2);
         W2 = (u16bit)(T ^ W3 ^ (16*j+k+1));
         W3 = W4; W4 = W1; W1 = T;
         }
      for(s32bit k = 7; k >= 0; k--)
         {
         u16bit T = W2; W2 = W3; W3 = W4;
         W4 = (u16bit)(W1 ^ T ^ (16*j+k+1));
         byte G1 = get_byte(1, T); byte G2 = get_byte(0, T);
         byte G3 = F[G2 ^ K[(4*(16*j+k)+3)%10]] ^ G1;
         G1 = F[G3 ^ K[(4*(16*j+k)+2)%10]] ^ G2;
         G2 = F[G1 ^ K[(4*(16*j+k)+1)%10]] ^ G3;
         G3 = F[G2 ^ K[(4*(16*j+k)  )%10]] ^ G1;
         W1 = make_u16bit(G3, G2);
         }
      }
   out[0] = get_byte(0, W1); out[1] = get_byte(1, W1);
   out[2] = get_byte(0, W2); out[3] = get_byte(1, W2);
   out[4] = get_byte(0, W3); out[5] = get_byte(1, W3);
   out[6] = get_byte(0, W4); out[7] = get_byte(1, W4);
   }

/*************************************************
* Skipjack Key Setup                             *
*************************************************/
void Skipjack::set_key(const byte key[], u32bit length) throw(InvalidKeyLength)
   {
   if(length != KEYLENGTH)
      throw InvalidKeyLength("Skipjack");
   K.copy(key, KEYLENGTH);
   }

}
