/*************************************************
* XTEA Source File                               *
* (C) 1999-2001 The OpenCL Project               *
*************************************************/

#include "xtea.h"

namespace OpenCL {

/*************************************************
* XTEA Encryption                                *
*************************************************/
void XTEA::encrypt(const byte in[BLOCKSIZE], byte out[BLOCKSIZE]) const
   {
   u32bit left  = make_u32bit(in[0], in[1], in[2], in[3]);
   u32bit right = make_u32bit(in[4], in[5], in[6], in[7]);
   u32bit sum = 0;
   for(u32bit j = 0; j != ROUNDS; j++)
      {
      left  += (((right << 4) ^ (right >> 5)) + right) ^ (sum + K[sum%4]);
      sum   += DELTA;
      right += (((left << 4) ^ (left >> 5)) + left) ^ (sum + K[(sum >> 11)%4]);
      }
   out[0] = get_byte(0, left);  out[1] = get_byte(1, left);
   out[2] = get_byte(2, left);  out[3] = get_byte(3, left);
   out[4] = get_byte(0, right); out[5] = get_byte(1, right);
   out[6] = get_byte(2, right); out[7] = get_byte(3, right);
   }

/*************************************************
* XTEA Decryption                                *
*************************************************/
void XTEA::decrypt(const byte in[BLOCKSIZE], byte out[BLOCKSIZE]) const
   {
   u32bit left  = make_u32bit(in[0], in[1], in[2], in[3]);
   u32bit right = make_u32bit(in[4], in[5], in[6], in[7]);
   u32bit sum = DELTA << LOG2_ROUNDS;
   for(u32bit j = 0; j != ROUNDS; j++)
      {
      right -= (((left << 4) ^ (left >> 5)) + left) ^ (sum + K[(sum >> 11)%4]);
      sum   -= DELTA;
      left  -= (((right << 4) ^ (right >> 5)) + right) ^ (sum + K[sum%4]);
      }
   out[0] = get_byte(0, left);  out[1] = get_byte(1, left);
   out[2] = get_byte(2, left);  out[3] = get_byte(3, left);
   out[4] = get_byte(0, right); out[5] = get_byte(1, right);
   out[6] = get_byte(2, right); out[7] = get_byte(3, right);
   }

/*************************************************
* XTEA Key Setup                                 *
*************************************************/
void XTEA::set_key(const byte key[], u32bit length) throw(InvalidKeyLength)
   {
   if(length != KEYLENGTH)
      throw InvalidKeyLength("XTEA");
   for(u32bit j = 0; j != 4; j++)
      K[j] = make_u32bit(key[4*j], key[4*j+1], key[4*j+2], key[4*j+3]);
   }

}
